
# this code is auto-generated by the script 'pyclesperanto_autogen_tier_script.ipynb'.
# Do not edit manually. Instead, edit the script and run it again.

from ._core import Device
from ._array import Image
from ._decorators import plugin_function


@plugin_function
def exclude_labels_on_edges(
    input_image: Image,
	output_image: Image = None,
	exclude_x: bool = True,
	exclude_y: bool = True,
	exclude_z: bool = True,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _exclude_labels_on_edges as op

    return op(
        device=device,
		src=input_image,
		dst=output_image,
		exclude_x=bool(exclude_x),
		exclude_y=bool(exclude_y),
		exclude_z=bool(exclude_z)
    )


@plugin_function
def flag_existing_labels(
    input_image: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _flag_existing_labels as op

    return op(
        device=device,
		src=input_image,
		dst=output_image
    )


@plugin_function
def gamma_correction(
    input_image: Image,
	output_image: Image = None,
	gamma: float = 0,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _gamma_correction as op

    return op(
        device=device,
		src=input_image,
		dst=output_image,
		gamma=float(gamma)
    )


@plugin_function
def histogram(
    input_image: Image,
	output_image: Image = None,
	nbins: int = 0,
	min: float = 0,
	max: float = 0,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _histogram as op

    return op(
        device=device,
		src=input_image,
		dst=output_image,
		nbins=int(nbins),
		min=float(min),
		max=float(max)
    )


@plugin_function
def mean_of_all_pixels(
    input_image: Image,
	device: Device = None
) -> float:
    from ._pyclesperanto import _mean_of_all_pixels as op

    return op(
        device=device,
		src=input_image
    )




from ._core import (
    select_backend,
    get_device,
    select_device,
    list_available_devices,
    list_available_backends,
    default_initialisation,
    Device,
    gpu_info,
)
from ._array import Image, is_image
from ._memory import create, create_like, push, pull
from ._functionalities import imshow, list_operations

from ._tier1 import *
from ._tier2 import *
from ._tier3 import *
from ._tier4 import *
from ._tier5 import *
from ._tier6 import *
from ._tier7 import *

from ._version import VERSION as __version__
from ._version import CLIC_VERSION as __clic_version__
from ._version import COMMON_ALIAS as __common_alias__

# Aliases - compatibility with prototype
from ._memory import push as asarray
from ._core import list_available_devices as available_device_names
from ._core import gpu_info as cl_info
from ._tier5 import connected_components_labeling_box as label

default_initialisation()




# this code is auto-generated by the script 'pyclesperanto_autogen_tier_script.ipynb'.
# Do not edit manually. Instead, edit the script and run it again.

from ._core import Device
from ._array import Image
from ._decorators import plugin_function


@plugin_function
def mean_squared_error(
    input_image0: Image,
	input_image1: Image,
	device: Device = None
) -> float:
    from ._pyclesperanto import _mean_squared_error as op

    return op(
        device=device,
		src0=input_image0,
		src1=input_image1
    )


@plugin_function
def relabel_sequential(
    input_image: Image,
	output_image: Image = None,
	blocksize: int = 0,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _relabel_sequential as op

    return op(
        device=device,
		src=input_image,
		dst=output_image,
		blocksize=int(blocksize)
    )


@plugin_function
def threshold_otsu(
    input_image: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _threshold_otsu as op

    return op(
        device=device,
		src=input_image,
		dst=output_image
    )




# from __future__ import annotations

from ._pyclesperanto import _Array as Array
from . import _operators

from typing import Union
import numpy as np
import warnings


def __str__(self) -> str:
    return self.get().__str__()


def __repr__(self) -> str:
    repr_str = self.get().__repr__()
    extra_info = f"mtype={self.mtype}"
    return repr_str[:-1] + f", {extra_info})"


def set(self, array: np.ndarray) -> None:
    if array.dtype != self.dtype:
        warnings.warn(
            f"Array dtype mismatch. Casting array to '{self.dtype.__name__}' before set().",
            RuntimeWarning,
        )
        array = array.astype(self.dtype)
    if array.size != self.size:
        raise ValueError(
            f"Array size mismatch: {array.size} != {self.size} ({array.shape} != {self.shape})"
        )
    if array.ndim != self.ndim:
        raise ValueError(
            f"Array dimension mismatch: {array.ndim} != {self.ndim} ({array.shape} != {self.shape})"
        )
    if array.shape != self.shape:
        raise ValueError(f"Array shape mismatch: {array.shape} != {self.shape}")
    self._write(array)
    return self


def get(self) -> np.ndarray:
    caster = {
        "float32": self._read_float32,
        "int8": self._read_int8,
        "int16": self._read_int16,
        "int32": self._read_int32,
        "int64": self._read_int64,
        "uint8": self._read_uint8,
        "uint16": self._read_uint16,
        "uint32": self._read_uint32,
        "uint64": self._read_uint64,
    }
    return caster[self.dtype.name]()


def __array__(self, dtype=None) -> np.ndarray:
    if dtype is None:
        return self.get()
    else:
        return self.get().astype(dtype)


# missing operators:
# __setitem__
# __getitem__
# __iter__
# __array_interface__

setattr(Array, "set", set)
setattr(Array, "get", get)
setattr(Array, "__str__", __str__)
setattr(Array, "__repr__", __repr__)
setattr(Array, "__array__", __array__)
setattr(Array, "astype", _operators.astype)
setattr(Array, "max", _operators.max)
setattr(Array, "min", _operators.min)
setattr(Array, "sum", _operators.sum)
setattr(Array, "__iadd__", _operators.__iadd__)
setattr(Array, "__sub__", _operators.__sub__)
setattr(Array, "__div__", _operators.__div__)
setattr(Array, "__truediv__", _operators.__truediv__)
setattr(Array, "__idiv__", _operators.__idiv__)
setattr(Array, "__itruediv__", _operators.__itruediv__)
setattr(Array, "__mul__", _operators.__mul__)
setattr(Array, "__imul__", _operators.__imul__)
setattr(Array, "__gt__", _operators.__gt__)
setattr(Array, "__ge__", _operators.__ge__)
setattr(Array, "__lt__", _operators.__lt__)
setattr(Array, "__le__", _operators.__le__)
setattr(Array, "__eq__", _operators.__eq__)
setattr(Array, "__ne__", _operators.__ne__)
setattr(Array, "__pow__", _operators.__pow__)
setattr(Array, "__ipow__", _operators.__ipow__)
setattr(Array, "_plt_to_png", _operators._plt_to_png)
setattr(Array, "_png_to_html", _operators._png_to_html)
setattr(Array, "_repr_html_", _operators._repr_html_)

Image = Union[np.ndarray, Array]


def is_image(any_array):
    return (
        isinstance(any_array, np.ndarray)
        or isinstance(any_array, tuple)
        or isinstance(any_array, list)
        or isinstance(any_array, Array)
        or str(type(any_array))
        in [
            "<class 'cupy._core.core.ndarray'>",
            "<class 'dask.array.core.Array'>",
            "<class 'xarray.core.dataarray.DataArray'>",
        ]
    )



from typing import Callable, Optional
from functools import wraps
from toolz import curry
import inspect

from ._memory import push
from ._array import is_image, Image
from ._core import Device


@curry
def plugin_function(
    function: Callable,
    categories: Optional[list] = None,
    priority: int = 0,
) -> Callable:
    """Function decorator to ensure correct types and values of all parameters.

    The given input parameters are either of type OpenCL data/image/buffer (which the GPU
    understands) or are converted to this type (see push function). If output
    parameters of type Image are not set, an empty image is created and
    handed over.

    Parameters
    ----------
    function : callable
        The function to be executed on the GPU.
    output_creator : callable, optional
        A function to create an output cleImage given an input cleImage. By
        default, we create output images of the same shape and type as input
        images.
    device_selector : callable, optional
        A function to select a device. By default, we use the current device instance.
    categories : list of str, optional
        A list of category names the function is associated with
    priority : int, optional
        can be used in lists of multiple operations to differentiate multiple operations that fulfill the same purpose
        but better/faster/more general.

    Returns
    -------
    worker_function : callable
        The actual function call that will be executed, magically creating
        output arguments of the correct type.
    """

    function.fullargspec = inspect.getfullargspec(function)
    function.categories = categories
    function.priority = priority

    @wraps(function)
    def worker_function(*args, **kwargs):
        sig = inspect.signature(function)
        # create mapping from position and keyword arguments to parameters
        # will raise a TypeError if the provided arguments do not match the signature
        # https://docs.python.org/3/library/inspect.html#inspect.Signature.bind
        bound = sig.bind(*args, **kwargs)
        # set default values for missing arguments
        # https://docs.python.org/3/library/inspect.html#inspect.BoundArguments.apply_defaults
        bound.apply_defaults()

        args_list = function.fullargspec.args
        index = next(
            (i for i, element in enumerate(args_list) if "input_image" in element), -1
        )
        arg_name = args_list[index]

        # copy images to GPU, and create output array if necessary
        for key, value in bound.arguments.items():
            if (
                is_image(value)
                and key in sig.parameters
                and sig.parameters[key].annotation is Image
            ):
                bound.arguments[key] = push(value)
            if (
                key in sig.parameters
                and sig.parameters[key].annotation is Device
                and value is None
            ):
                input_image = bound.arguments[arg_name]
                bound.arguments[key] = input_image.device

        # call the decorated function
        result = function(*bound.args, **bound.kwargs)

        # # Cast the result as an Array if it is not already
        # if not isinstance(result, _Array):
        #     result = _Array(result)

        return result

    # this is necessary to obfuscate pyclesperanto's internal structure
    worker_function.__module__ = "pyclesperanto"

    return worker_function




# this code is auto-generated by the script 'pyclesperanto_autogen_tier_script.ipynb'.
# Do not edit manually. Instead, edit the script and run it again.

from ._core import Device
from ._array import Image
from ._decorators import plugin_function


@plugin_function
def absolute_difference(
    input_image0: Image,
	input_image1: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _absolute_difference as op

    return op(
        device=device,
		src0=input_image0,
		src1=input_image1,
		dst=output_image
    )


@plugin_function
def add_images(
    input_image0: Image,
	input_image1: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _add_images as op

    return op(
        device=device,
		src0=input_image0,
		src1=input_image1,
		dst=output_image
    )


@plugin_function
def bottom_hat_box(
    input_image: Image,
	output_image: Image = None,
	radius_x: int = 0,
	radius_y: int = 0,
	radius_z: int = 0,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _bottom_hat_box as op

    return op(
        device=device,
		src=input_image,
		dst=output_image,
		radius_x=int(radius_x),
		radius_y=int(radius_y),
		radius_z=int(radius_z)
    )


@plugin_function
def bottom_hat_sphere(
    input_image: Image,
	output_image: Image = None,
	radius_x: float = 0,
	radius_y: float = 0,
	radius_z: float = 0,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _bottom_hat_sphere as op

    return op(
        device=device,
		src=input_image,
		dst=output_image,
		radius_x=float(radius_x),
		radius_y=float(radius_y),
		radius_z=float(radius_z)
    )


@plugin_function
def clip(
    input_image: Image,
	output_image: Image = None,
	min_intensity: float = 0,
	max_intensity: float = 0,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _clip as op

    return op(
        device=device,
		src=input_image,
		dst=output_image,
		min_intensity=float(min_intensity),
		max_intensity=float(max_intensity)
    )


@plugin_function
def closing_box(
    input_image: Image,
	output_image: Image = None,
	radius_x: int = 0,
	radius_y: int = 0,
	radius_z: int = 0,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _closing_box as op

    return op(
        device=device,
		src=input_image,
		dst=output_image,
		radius_x=int(radius_x),
		radius_y=int(radius_y),
		radius_z=int(radius_z)
    )


@plugin_function
def closing_sphere(
    input_image: Image,
	output_image: Image = None,
	radius_x: float = 0,
	radius_y: float = 0,
	radius_z: float = 0,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _closing_sphere as op

    return op(
        device=device,
		src=input_image,
		dst=output_image,
		radius_x=float(radius_x),
		radius_y=float(radius_y),
		radius_z=float(radius_z)
    )


@plugin_function
def combine_horizontally(
    input_image0: Image,
	input_image1: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _combine_horizontally as op

    return op(
        device=device,
		src0=input_image0,
		src1=input_image1,
		dst=output_image
    )


@plugin_function
def combine_vertically(
    input_image0: Image,
	input_image1: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _combine_vertically as op

    return op(
        device=device,
		src0=input_image0,
		src1=input_image1,
		dst=output_image
    )


@plugin_function
def concatenate_stacks(
    input_image0: Image,
	input_image1: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _concatenate_stacks as op

    return op(
        device=device,
		src0=input_image0,
		src1=input_image1,
		dst=output_image
    )


@plugin_function
def degrees_to_radians(
    input_image: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _degrees_to_radians as op

    return op(
        device=device,
		src=input_image,
		dst=output_image
    )


@plugin_function
def difference_of_gaussian(
    input_image: Image,
	output_image: Image = None,
	sigma1_x: float = 0,
	sigma1_y: float = 0,
	sigma1_z: float = 0,
	sigma2_x: float = 0,
	sigma2_y: float = 0,
	sigma2_z: float = 0,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _difference_of_gaussian as op

    return op(
        device=device,
		src=input_image,
		dst=output_image,
		sigma1_x=float(sigma1_x),
		sigma1_y=float(sigma1_y),
		sigma1_z=float(sigma1_z),
		sigma2_x=float(sigma2_x),
		sigma2_y=float(sigma2_y),
		sigma2_z=float(sigma2_z)
    )


@plugin_function
def extend_labeling_via_voronoi(
    input_image: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _extend_labeling_via_voronoi as op

    return op(
        device=device,
		src=input_image,
		dst=output_image
    )


@plugin_function
def invert(
    input_image: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _invert as op

    return op(
        device=device,
		src=input_image,
		dst=output_image
    )


@plugin_function
def label_spots(
    input_image: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _label_spots as op

    return op(
        device=device,
		src=input_image,
		dst=output_image
    )


@plugin_function
def maximum_of_all_pixels(
    input_image: Image,
	device: Device = None
) -> float:
    from ._pyclesperanto import _maximum_of_all_pixels as op

    return op(
        device=device,
		src=input_image
    )


@plugin_function
def minimum_of_all_pixels(
    input_image: Image,
	device: Device = None
) -> float:
    from ._pyclesperanto import _minimum_of_all_pixels as op

    return op(
        device=device,
		src=input_image
    )


@plugin_function
def opening_box(
    input_image: Image,
	output_image: Image = None,
	radius_x: int = 0,
	radius_y: int = 0,
	radius_z: int = 0,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _opening_box as op

    return op(
        device=device,
		src=input_image,
		dst=output_image,
		radius_x=int(radius_x),
		radius_y=int(radius_y),
		radius_z=int(radius_z)
    )


@plugin_function
def opening_sphere(
    input_image: Image,
	output_image: Image = None,
	radius_x: float = 0,
	radius_y: float = 0,
	radius_z: float = 0,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _opening_sphere as op

    return op(
        device=device,
		src=input_image,
		dst=output_image,
		radius_x=float(radius_x),
		radius_y=float(radius_y),
		radius_z=float(radius_z)
    )


@plugin_function
def radians_to_degrees(
    input_image: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _radians_to_degrees as op

    return op(
        device=device,
		src=input_image,
		dst=output_image
    )


@plugin_function
def square(
    input_image: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _square as op

    return op(
        device=device,
		src=input_image,
		dst=output_image
    )


@plugin_function
def squared_difference(
    input_image0: Image,
	input_image1: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _squared_difference as op

    return op(
        device=device,
		src0=input_image0,
		src1=input_image1,
		dst=output_image
    )


@plugin_function
def subtract_images(
    input_image0: Image,
	input_image1: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _subtract_images as op

    return op(
        device=device,
		src0=input_image0,
		src1=input_image1,
		dst=output_image
    )


@plugin_function
def sum_of_all_pixels(
    input_image: Image,
	device: Device = None
) -> float:
    from ._pyclesperanto import _sum_of_all_pixels as op

    return op(
        device=device,
		src=input_image
    )


@plugin_function
def top_hat_box(
    input_image: Image,
	output_image: Image = None,
	radius_x: int = 0,
	radius_y: int = 0,
	radius_z: int = 0,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _top_hat_box as op

    return op(
        device=device,
		src=input_image,
		dst=output_image,
		radius_x=int(radius_x),
		radius_y=int(radius_y),
		radius_z=int(radius_z)
    )


@plugin_function
def top_hat_sphere(
    input_image: Image,
	output_image: Image = None,
	radius_x: float = 0,
	radius_y: float = 0,
	radius_z: float = 0,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _top_hat_sphere as op

    return op(
        device=device,
		src=input_image,
		dst=output_image,
		radius_x=float(radius_x),
		radius_y=float(radius_y),
		radius_z=float(radius_z)
    )




from typing import Optional
import warnings

from ._pyclesperanto import _Device as Device
from ._pyclesperanto import _BackendManager as BackendManager


class _current_device:
    _instance: Optional[Device] = None


def get_device() -> Device:
    """Return the current device instance

    Returns
    -------
    device : Device
    """
    return _current_device._instance or select_device()


def select_device(device_name: str = "", device_type: str = "all") -> Device:
    """Select a device by name and type, and return it

    Select device using its name or subname (e.g. "NVIDIA", "RTX", "Iris", etc.) and
    type (e.g. "all", "cpu", "gpu").
    To retrieve a list of available devices, use `list_available_devices()`

    Parameters
    ----------
    device_name : str, default = ""
        Name or subname of the device to be selected (e.g. "NVIDIA", "RTX", "Intel Iris", etc.)
    device_type : str, default = "all"
        Type of device to be selected (e.g. "all", "cpu", "gpu")

    Returns
    -------
    device : Device
    """
    device = BackendManager.get_backend().getDevice(device_name, device_type)
    if _current_device._instance and device == _current_device._instance:
        return _current_device._instance
    _current_device._instance = device
    return device


def list_available_devices(device_type: str = "all") -> list:
    """Retrieve a list of names of available devices

    Will search system for backend available compatible device and return a list of their names.
    Use 'select_device' or 'get_devices' to select devices.

    Parameters
    ----------
    device_type : str, default = "all"
        Type of device to be selected (e.g. "all", "cpu", "gpu")

    Returns
    -------
    name list : list[str]
    """
    dev_list = list(BackendManager.get_backend().getDevicesList(type=device_type))
    if not dev_list:
        warnings.warn(
            "No device available. Please install either OpenCL or CUDA on your system.",
            RuntimeWarning,
        )
    return dev_list


def list_available_backends() -> list:
    """Retrieve a list of names of available backends

    Will test system for available backends and return a list of their names.

    Returns
    -------
    name list : list[str]
    """
    back_list = list(BackendManager.get_backends_list())
    if not back_list:
        warnings.warn(
            "No backend available. Please install either OpenCL or CUDA on your system.",
            RuntimeWarning,
        )
    return back_list


def select_backend(backend: str = "opencl") -> str:
    """select backend

    Select the backend used by pyclesperanto, OpenCL or CUDA.
    Default is OpenCL.

    Parameters
    ----------
    type : str, default = "opencl"
        determine the backend to use between opencl and cuda
    """

    # enforce lowercase for backend_type
    backend = backend.lower()
    # is backend_type is different than "cuda" or "opencl", raise an error
    if backend not in ["cuda", "opencl"]:
        raise ValueError(
            f"'{backend}' is not a supported Backend. Please use either 'opencl' or 'cuda'."
        )
    BackendManager.set_backend(backend=backend)
    # reset current device to default one
    select_device()
    return f"{BackendManager.get_backend()} selected."


def wait_for_kernel_to_finish(flag: bool = True, device: Device = None):
    """Wait for kernel to finish

    Parameters
    ----------
    flag : bool, default = True
        if True, wait for kernel to finish
    """
    if device is None:
        get_device().set_wait_to_finish(flag)
    else:
        device.set_wait_to_finish(flag)


def default_initialisation():
    backends = list_available_backends()
    if backends:
        _ = select_backend(backends[-1])
    else:
        warnings.warn(
            "No backend available. Please install either OpenCL or CUDA on your system.",
            RuntimeWarning,
        )


def gpu_info():
    device_list = list_available_devices("gpu")
    for device_name in device_list:
        print(select_device(device_name).info)
        print("")



from os import path
from typing import Optional, Union

from ._array import Array, Image
from ._memory import pull
from ._core import Device, get_device

# def execute(
#     anchor: str,
#     kernel_filepath: str,
#     kernel_name: str,
#     parameters: dict,
#     range: Optional[tuple] = None,
#     device: Optional[Device] = None,
# ):
#     """Execute a custom OpenCL kernel.

#     Parameters
#     ----------
#     anchor : str
#         Path to the directory where the kernel file is located.
#     opencl_kernel_filename : str
#         Name of the OpenCL kernel file, e.g. "my_kernel.cl".
#     kernel_name : str
#         Name of the kernel function to be executed in the OpenCL kernel file.
#     parameters : dict
#         Dictionary of parameters to be passed to the kernel function, e.g. {"src": src, "dst": dst}.
#     range : tuple, optional
#         Global size of the kernel execution, by default None.
#     device : Device, optional
#         Device to be used for execution, by default None.
#     """
#     from ._pyclesperanto import _std_variant as std_variant
#     from ._pyclesperanto import _execute as op

#     if range is None:
#         range = (1, 1, 1)
#     else:
#         if len(range) == 2:
#             range = (1, range[0], range[1])
#         if len(range) == 1:
#             range = (1, 1, range[0])

#     if device is None:
#         device = parameters["src"].device or parameters["src1"].device or get_device()

#     kernel_source = open(path.join(anchor, kernel_filepath), "r").read()

#     cpp_parameter_map = {
#         key: std_variant(val.super) if isinstance(val, Array) else std_variant(val)
#         for key, val in parameters.items()
# }
# op(
#     device,
#     kernel_name=kernel_name,
#     kernel_source=kernel_source,
#     parameters=cpp_parameter_map,
#     range=range,
#     # constants=cpp_constant_map,
# )


def imshow(
    image,
    title: Optional[str] = None,
    labels: Optional[bool] = False,
    min_display_intensity: Optional[float] = None,
    max_display_intensity: Optional[float] = None,
    color_map: Optional[str] = None,
    plot=None,
    colorbar: Optional[bool] = False,
    colormap: Union[str, None] = None,
    alpha: Optional[float] = None,
    continue_drawing: Optional[bool] = False,
):
    """Visualize an image, e.g. in Jupyter notebooks.
    Parameters
    ----------
    image: np.ndarray
        numpy or OpenCL-backed image to visualize
    title: str, optional
        Obsolete (kept for ImageJ-compatibility)
    labels: bool, optional
        True: integer labels will be visualized with colors
        False: Specified or default colormap will be used to display intensities.
    min_display_intensity: float, optional
        lower limit for display range
    max_display_intensity: float, optional
        upper limit for display range
    color_map: str, optional
        deprecated, use colormap instead
    plot: matplotlib axis, optional
        Plot object where the image should be shown. Useful for putting multiple images in subfigures.
    colorbar: bool, optional
        True puts a colorbar next to the image. Will not work with label images and when visualizing multiple
        images (continue_drawing=True).
    colormap: str or matplotlib colormap, optional
    alpha: float, optional
        alpha blending value
    continue_drawing: float
        True: the next shown image can be visualized on top of the current one, e.g. with alpha = 0.5
    """
    if len(image.shape) == 3:
        from ._tier1 import maximum_z_projection

        image = pull(maximum_z_projection(image))

    image = pull(image)

    if color_map is not None:
        import warnings

        warnings.warn(
            "The imshow parameter color_map is deprecated. Use colormap instead."
        )
        if colormap is None:
            colormap = color_map

    if colormap is None:
        colormap = "Greys_r"

    cmap = colormap
    if labels:
        if not hasattr(imshow, "labels_cmap"):
            from matplotlib.colors import ListedColormap
            from numpy.random import MT19937
            from numpy.random import RandomState, SeedSequence

            rs = RandomState(MT19937(SeedSequence(3)))
            lut = rs.rand(65537, 3)
            lut[0, :] = 0
            # these are the first four colours from matplotlib's default
            lut[1] = [0.12156862745098039, 0.4666666666666667, 0.7058823529411765]
            lut[2] = [1.0, 0.4980392156862745, 0.054901960784313725]
            lut[3] = [0.17254901960784313, 0.6274509803921569, 0.17254901960784313]
            lut[4] = [0.8392156862745098, 0.15294117647058825, 0.1568627450980392]
            imshow.labels_cmap = ListedColormap(lut)

        cmap = imshow.labels_cmap
        if min_display_intensity is None:
            min_display_intensity = 0
        if max_display_intensity is None:
            max_display_intensity = 65536

    if plot is None:
        import matplotlib.pyplot as plt

        plt.imshow(
            image,
            cmap=cmap,
            vmin=min_display_intensity,
            vmax=max_display_intensity,
            interpolation="nearest",
            alpha=alpha,
        )
        if colorbar:
            plt.colorbar()
        if not continue_drawing:
            plt.show()
    else:
        plot.imshow(
            image,
            cmap=cmap,
            vmin=min_display_intensity,
            vmax=max_display_intensity,
            interpolation="nearest",
            alpha=alpha,
        )
        if colorbar:
            plot.colorbar()
    if title is not None:
        plt.title(title)


def operations(
    must_have_categories: list = None, must_not_have_categories: list = None
) -> dict:
    """Retrieve a dictionary of operations, which can be filtered by annotated categories.

    Parameters
    ----------
    must_have_categories : list of str, optional
        if provided, the result will be filtered so that operations must contain all given categories.
    must_not_have_categories : list of str, optional
        if provided, the result will be filtered so that operations must not contain all given categories.

    Returns
    -------
    dict of str : function
    """
    if isinstance(must_have_categories, str):
        must_have_categories = [must_have_categories]
    if isinstance(must_not_have_categories, str):
        must_have_categories = [must_not_have_categories]

    result = {}

    from inspect import getmembers, isfunction
    import pyclesperanto as cle

    # retrieve all operations and cache the result for later reuse
    if not hasattr(operations, "_all") or operations._all is None:
        operations._all = getmembers(cle, isfunction)

    # filter operations according to given constraints
    for operation_name, operation in operations._all:
        keep_it = True
        if hasattr(operation, "categories") and operation.categories is not None:
            if must_have_categories is not None:
                if not all(
                    item in operation.categories for item in must_have_categories
                ):
                    keep_it = False

            if must_not_have_categories is not None:
                if any(
                    item in operation.categories for item in must_not_have_categories
                ):
                    keep_it = False
        else:
            if must_have_categories is not None:
                keep_it = False
        if keep_it:
            result[operation_name] = operation

    return result


def list_operations(search_term=None):
    ops = operations(search_term)
    for name in ops:
        func = ops[name]
        if hasattr(func, "fullargspec"):
            print(
                name
                + "("
                + str(func.fullargspec.args)
                .replace("[", "")
                .replace("]", "")
                .replace("'", "")
                + ")"
            )
        else:
            print(name)



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import os
import sys
sys.path.insert(0, os.path.abspath('../../'))

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'pyclesperanto'
copyright = '2023, Stephane Rigaud'
author = 'Stephane Rigaud'
release = '0.6.5'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration


extensions = [
  'sphinx.ext.autodoc',  # Parses (sub)modules
  'sphinx.ext.napoleon',  # Parses Numpy docstrings
  'sphinx.ext.mathjax',  # Print mathematical expressions
  'nbsphinx',  # link notebooks
  'sphinx.ext.autosummary',  # Make module lists in table
]

templates_path = ['_templates']
exclude_patterns = ['build', '_build', 'Thumbs.db', '.DS_Store']

autodoc_mock_imports = ["pyclesperanto._pyclesperanto", "toolz", "matplotlib", "numpy"]
add_module_names = False
modindex_common_prefix = ['pyclesperanto.']

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'sphinx_rtd_theme' # 'alabaster'
html_static_path = ['_static']




# this code is auto-generated by the script 'pyclesperanto_autogen_tier_script.ipynb'.
# Do not edit manually. Instead, edit the script and run it again.

from ._core import Device
from ._array import Image
from ._decorators import plugin_function


@plugin_function
def absolute(
    input_image: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _absolute as op

    return op(
        device=device,
		src=input_image,
		dst=output_image
    )


@plugin_function
def add_images_weighted(
    input_image0: Image,
	input_image1: Image,
	output_image: Image = None,
	factor0: float = 0,
	factor1: float = 0,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _add_images_weighted as op

    return op(
        device=device,
		src0=input_image0,
		src1=input_image1,
		dst=output_image,
		factor0=float(factor0),
		factor1=float(factor1)
    )


@plugin_function
def add_image_and_scalar(
    input_image: Image,
	output_image: Image = None,
	scalar: float = 0,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _add_image_and_scalar as op

    return op(
        device=device,
		src=input_image,
		dst=output_image,
		scalar=float(scalar)
    )


@plugin_function
def binary_and(
    input_image0: Image,
	input_image1: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _binary_and as op

    return op(
        device=device,
		src0=input_image0,
		src1=input_image1,
		dst=output_image
    )


@plugin_function
def binary_edge_detection(
    input_image: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _binary_edge_detection as op

    return op(
        device=device,
		src=input_image,
		dst=output_image
    )


@plugin_function
def binary_not(
    input_image: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _binary_not as op

    return op(
        device=device,
		src=input_image,
		dst=output_image
    )


@plugin_function
def binary_or(
    input_image0: Image,
	input_image1: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _binary_or as op

    return op(
        device=device,
		src0=input_image0,
		src1=input_image1,
		dst=output_image
    )


@plugin_function
def binary_subtract(
    input_image0: Image,
	input_image1: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _binary_subtract as op

    return op(
        device=device,
		src0=input_image0,
		src1=input_image1,
		dst=output_image
    )


@plugin_function
def binary_xor(
    input_image0: Image,
	input_image1: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _binary_xor as op

    return op(
        device=device,
		src0=input_image0,
		src1=input_image1,
		dst=output_image
    )


@plugin_function
def block_enumerate(
    input_image0: Image,
	input_image1: Image,
	output_image: Image = None,
	blocksize: int = 0,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _block_enumerate as op

    return op(
        device=device,
		src0=input_image0,
		src1=input_image1,
		dst=output_image,
		blocksize=int(blocksize)
    )


@plugin_function
def convolve(
    input_image0: Image,
	input_image1: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _convolve as op

    return op(
        device=device,
		src0=input_image0,
		src1=input_image1,
		dst=output_image
    )


@plugin_function
def copy(
    input_image: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _copy as op

    return op(
        device=device,
		src=input_image,
		dst=output_image
    )


@plugin_function
def copy_slice(
    input_image: Image,
	output_image: Image = None,
	slice: int = 0,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _copy_slice as op

    return op(
        device=device,
		src=input_image,
		dst=output_image,
		slice=int(slice)
    )


@plugin_function
def copy_horizontal_slice(
    input_image: Image,
	output_image: Image = None,
	slice: int = 0,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _copy_horizontal_slice as op

    return op(
        device=device,
		src=input_image,
		dst=output_image,
		slice=int(slice)
    )


@plugin_function
def copy_vertical_slice(
    input_image: Image,
	output_image: Image = None,
	slice: int = 0,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _copy_vertical_slice as op

    return op(
        device=device,
		src=input_image,
		dst=output_image,
		slice=int(slice)
    )


@plugin_function
def crop(
    input_image: Image,
	output_image: Image = None,
	start_x: int = 0,
	start_y: int = 0,
	start_z: int = 0,
	width: int = 0,
	height: int = 0,
	depth: int = 0,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _crop as op

    return op(
        device=device,
		src=input_image,
		dst=output_image,
		start_x=int(start_x),
		start_y=int(start_y),
		start_z=int(start_z),
		width=int(width),
		height=int(height),
		depth=int(depth)
    )


@plugin_function
def cubic_root(
    input_image: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _cubic_root as op

    return op(
        device=device,
		src=input_image,
		dst=output_image
    )


@plugin_function
def detect_label_edges(
    input_image: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _detect_label_edges as op

    return op(
        device=device,
		src=input_image,
		dst=output_image
    )


@plugin_function
def detect_maxima_box(
    input_image: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _detect_maxima_box as op

    return op(
        device=device,
		src=input_image,
		dst=output_image
    )


@plugin_function
def detect_minima_box(
    input_image: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _detect_minima_box as op

    return op(
        device=device,
		src=input_image,
		dst=output_image
    )


@plugin_function
def dilate_box(
    input_image: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _dilate_box as op

    return op(
        device=device,
		src=input_image,
		dst=output_image
    )


@plugin_function
def dilate_sphere(
    input_image: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _dilate_sphere as op

    return op(
        device=device,
		src=input_image,
		dst=output_image
    )


@plugin_function
def divide_images(
    input_image0: Image,
	input_image1: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _divide_images as op

    return op(
        device=device,
		src0=input_image0,
		src1=input_image1,
		dst=output_image
    )


@plugin_function
def divide_image_and_scalar(
    input_image: Image,
	output_image: Image = None,
	scalar: float = 0,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _divide_image_and_scalar as op

    return op(
        device=device,
		src=input_image,
		dst=output_image,
		scalar=float(scalar)
    )


@plugin_function
def equal(
    input_image0: Image,
	input_image1: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _equal as op

    return op(
        device=device,
		src0=input_image0,
		src1=input_image1,
		dst=output_image
    )


@plugin_function
def equal_constant(
    input_image: Image,
	output_image: Image = None,
	scalar: float = 0,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _equal_constant as op

    return op(
        device=device,
		src=input_image,
		dst=output_image,
		scalar=float(scalar)
    )


@plugin_function
def erode_box(
    input_image: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _erode_box as op

    return op(
        device=device,
		src=input_image,
		dst=output_image
    )


@plugin_function
def erode_sphere(
    input_image: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _erode_sphere as op

    return op(
        device=device,
		src=input_image,
		dst=output_image
    )


@plugin_function
def exponential(
    input_image: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _exponential as op

    return op(
        device=device,
		src=input_image,
		dst=output_image
    )


@plugin_function
def flip(
    input_image: Image,
	output_image: Image = None,
	flip_x: bool = True,
	flip_y: bool = True,
	flip_z: bool = True,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _flip as op

    return op(
        device=device,
		src=input_image,
		dst=output_image,
		flip_x=bool(flip_x),
		flip_y=bool(flip_y),
		flip_z=bool(flip_z)
    )


@plugin_function
def gaussian_blur(
    input_image: Image,
	output_image: Image = None,
	sigma_x: float = 0,
	sigma_y: float = 0,
	sigma_z: float = 0,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _gaussian_blur as op

    return op(
        device=device,
		src=input_image,
		dst=output_image,
		sigma_x=float(sigma_x),
		sigma_y=float(sigma_y),
		sigma_z=float(sigma_z)
    )


@plugin_function
def gradient_x(
    input_image: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _gradient_x as op

    return op(
        device=device,
		src=input_image,
		dst=output_image
    )


@plugin_function
def gradient_y(
    input_image: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _gradient_y as op

    return op(
        device=device,
		src=input_image,
		dst=output_image
    )


@plugin_function
def gradient_z(
    input_image: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _gradient_z as op

    return op(
        device=device,
		src=input_image,
		dst=output_image
    )


@plugin_function
def greater(
    input_image0: Image,
	input_image1: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _greater as op

    return op(
        device=device,
		src0=input_image0,
		src1=input_image1,
		dst=output_image
    )


@plugin_function
def greater_constant(
    input_image: Image,
	output_image: Image = None,
	scalar: float = 0,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _greater_constant as op

    return op(
        device=device,
		src=input_image,
		dst=output_image,
		scalar=float(scalar)
    )


@plugin_function
def greater_or_equal(
    input_image0: Image,
	input_image1: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _greater_or_equal as op

    return op(
        device=device,
		src0=input_image0,
		src1=input_image1,
		dst=output_image
    )


@plugin_function
def greater_or_equal_constant(
    input_image: Image,
	output_image: Image = None,
	scalar: float = 0,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _greater_or_equal_constant as op

    return op(
        device=device,
		src=input_image,
		dst=output_image,
		scalar=float(scalar)
    )


@plugin_function
def laplace_box(
    input_image: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _laplace_box as op

    return op(
        device=device,
		src=input_image,
		dst=output_image
    )


@plugin_function
def laplace_diamond(
    input_image: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _laplace_diamond as op

    return op(
        device=device,
		src=input_image,
		dst=output_image
    )


@plugin_function
def logarithm(
    input_image: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _logarithm as op

    return op(
        device=device,
		src=input_image,
		dst=output_image
    )


@plugin_function
def mask(
    input_image: Image,
	mask: Image = None,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _mask as op

    return op(
        device=device,
		src=input_image,
		mask=mask,
		dst=output_image
    )


@plugin_function
def mask_label(
    input_image0: Image,
	input_image1: Image,
	output_image: Image = None,
	label: float = 0,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _mask_label as op

    return op(
        device=device,
		src0=input_image0,
		src1=input_image1,
		dst=output_image,
		label=float(label)
    )


@plugin_function
def maximum_image_and_scalar(
    input_image: Image,
	output_image: Image = None,
	scalar: float = 0,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _maximum_image_and_scalar as op

    return op(
        device=device,
		src=input_image,
		dst=output_image,
		scalar=float(scalar)
    )


@plugin_function
def maximum_images(
    input_image0: Image,
	input_image1: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _maximum_images as op

    return op(
        device=device,
		src0=input_image0,
		src1=input_image1,
		dst=output_image
    )


@plugin_function
def maximum_box(
    input_image: Image,
	output_image: Image = None,
	radius_x: int = 0,
	radius_y: int = 0,
	radius_z: int = 0,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _maximum_box as op

    return op(
        device=device,
		src=input_image,
		dst=output_image,
		radius_x=int(radius_x),
		radius_y=int(radius_y),
		radius_z=int(radius_z)
    )


@plugin_function
def maximum_x_projection(
    input_image: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _maximum_x_projection as op

    return op(
        device=device,
		src=input_image,
		dst=output_image
    )


@plugin_function
def maximum_y_projection(
    input_image: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _maximum_y_projection as op

    return op(
        device=device,
		src=input_image,
		dst=output_image
    )


@plugin_function
def maximum_z_projection(
    input_image: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _maximum_z_projection as op

    return op(
        device=device,
		src=input_image,
		dst=output_image
    )


@plugin_function
def mean_box(
    input_image: Image,
	output_image: Image = None,
	radius_x: int = 0,
	radius_y: int = 0,
	radius_z: int = 0,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _mean_box as op

    return op(
        device=device,
		src=input_image,
		dst=output_image,
		radius_x=int(radius_x),
		radius_y=int(radius_y),
		radius_z=int(radius_z)
    )


@plugin_function
def mean_sphere(
    input_image: Image,
	output_image: Image = None,
	radius_x: int = 0,
	radius_y: int = 0,
	radius_z: int = 0,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _mean_sphere as op

    return op(
        device=device,
		src=input_image,
		dst=output_image,
		radius_x=int(radius_x),
		radius_y=int(radius_y),
		radius_z=int(radius_z)
    )


@plugin_function
def mean_x_projection(
    input_image: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _mean_x_projection as op

    return op(
        device=device,
		src=input_image,
		dst=output_image
    )


@plugin_function
def mean_y_projection(
    input_image: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _mean_y_projection as op

    return op(
        device=device,
		src=input_image,
		dst=output_image
    )


@plugin_function
def mean_z_projection(
    input_image: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _mean_z_projection as op

    return op(
        device=device,
		src=input_image,
		dst=output_image
    )


@plugin_function
def median_box(
    input_image: Image,
	output_image: Image = None,
	radius_x: int = 0,
	radius_y: int = 0,
	radius_z: int = 0,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _median_box as op

    return op(
        device=device,
		src=input_image,
		dst=output_image,
		radius_x=int(radius_x),
		radius_y=int(radius_y),
		radius_z=int(radius_z)
    )


@plugin_function
def median_sphere(
    input_image: Image,
	output_image: Image = None,
	radius_x: int = 0,
	radius_y: int = 0,
	radius_z: int = 0,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _median_sphere as op

    return op(
        device=device,
		src=input_image,
		dst=output_image,
		radius_x=int(radius_x),
		radius_y=int(radius_y),
		radius_z=int(radius_z)
    )


@plugin_function
def minimum_box(
    input_image: Image,
	output_image: Image = None,
	radius_x: int = 0,
	radius_y: int = 0,
	radius_z: int = 0,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _minimum_box as op

    return op(
        device=device,
		src=input_image,
		dst=output_image,
		radius_x=int(radius_x),
		radius_y=int(radius_y),
		radius_z=int(radius_z)
    )


@plugin_function
def minimum_image_and_scalar(
    input_image: Image,
	output_image: Image = None,
	scalar: float = 0,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _minimum_image_and_scalar as op

    return op(
        device=device,
		src=input_image,
		dst=output_image,
		scalar=float(scalar)
    )


@plugin_function
def minimum_images(
    input_image0: Image,
	input_image1: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _minimum_images as op

    return op(
        device=device,
		src0=input_image0,
		src1=input_image1,
		dst=output_image
    )


@plugin_function
def minimum_x_projection(
    input_image: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _minimum_x_projection as op

    return op(
        device=device,
		src=input_image,
		dst=output_image
    )


@plugin_function
def minimum_y_projection(
    input_image: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _minimum_y_projection as op

    return op(
        device=device,
		src=input_image,
		dst=output_image
    )


@plugin_function
def minimum_z_projection(
    input_image: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _minimum_z_projection as op

    return op(
        device=device,
		src=input_image,
		dst=output_image
    )


@plugin_function
def mode_box(
    input_image: Image,
	output_image: Image = None,
	radius_x: int = 0,
	radius_y: int = 0,
	radius_z: int = 0,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _mode_box as op

    return op(
        device=device,
		src=input_image,
		dst=output_image,
		radius_x=int(radius_x),
		radius_y=int(radius_y),
		radius_z=int(radius_z)
    )


@plugin_function
def mode_sphere(
    input_image: Image,
	output_image: Image = None,
	radius_x: int = 0,
	radius_y: int = 0,
	radius_z: int = 0,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _mode_sphere as op

    return op(
        device=device,
		src=input_image,
		dst=output_image,
		radius_x=int(radius_x),
		radius_y=int(radius_y),
		radius_z=int(radius_z)
    )


@plugin_function
def modulo_images(
    input_image0: Image,
	input_image1: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _modulo_images as op

    return op(
        device=device,
		src0=input_image0,
		src1=input_image1,
		dst=output_image
    )


@plugin_function
def multiply_image_and_scalar(
    input_image: Image,
	output_image: Image = None,
	scalar: float = 0,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _multiply_image_and_scalar as op

    return op(
        device=device,
		src=input_image,
		dst=output_image,
		scalar=float(scalar)
    )


@plugin_function
def multiply_images(
    input_image0: Image,
	input_image1: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _multiply_images as op

    return op(
        device=device,
		src0=input_image0,
		src1=input_image1,
		dst=output_image
    )


@plugin_function
def nan_to_num(
    input_image: Image,
	output_image: Image = None,
	nan: float = 0,
	posinf: float = 0,
	neginf: float = 0,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _nan_to_num as op

    return op(
        device=device,
		src=input_image,
		dst=output_image,
		nan=float(nan),
		posinf=float(posinf),
		neginf=float(neginf)
    )


@plugin_function
def nonzero_maximum_box(
    input_image: Image,
	output_image0: Image = None,
	output_image1: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _nonzero_maximum_box as op

    return op(
        device=device,
		src=input_image,
		dst0=output_image0,
		dst1=output_image1
    )


@plugin_function
def nonzero_maximum_diamond(
    input_image: Image,
	output_image0: Image = None,
	output_image1: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _nonzero_maximum_diamond as op

    return op(
        device=device,
		src=input_image,
		dst0=output_image0,
		dst1=output_image1
    )


@plugin_function
def nonzero_minimum_box(
    input_image: Image,
	output_image0: Image = None,
	output_image1: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _nonzero_minimum_box as op

    return op(
        device=device,
		src=input_image,
		dst0=output_image0,
		dst1=output_image1
    )


@plugin_function
def nonzero_minimum_diamond(
    input_image: Image,
	output_image0: Image = None,
	output_image1: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _nonzero_minimum_diamond as op

    return op(
        device=device,
		src=input_image,
		dst0=output_image0,
		dst1=output_image1
    )


@plugin_function
def not_equal(
    input_image0: Image,
	input_image1: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _not_equal as op

    return op(
        device=device,
		src0=input_image0,
		src1=input_image1,
		dst=output_image
    )


@plugin_function
def not_equal_constant(
    input_image: Image,
	output_image: Image = None,
	scalar: float = 0,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _not_equal_constant as op

    return op(
        device=device,
		src=input_image,
		dst=output_image,
		scalar=float(scalar)
    )


@plugin_function
def paste(
    input_image: Image,
	output_image: Image = None,
	index_x: int = 0,
	index_y: int = 0,
	index_z: int = 0,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _paste as op

    return op(
        device=device,
		src=input_image,
		dst=output_image,
		index_x=int(index_x),
		index_y=int(index_y),
		index_z=int(index_z)
    )


@plugin_function
def onlyzero_overwrite_maximum_box(
    input_image: Image,
	output_image0: Image = None,
	output_image1: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _onlyzero_overwrite_maximum_box as op

    return op(
        device=device,
		src=input_image,
		dst0=output_image0,
		dst1=output_image1
    )


@plugin_function
def onlyzero_overwrite_maximum_diamond(
    input_image: Image,
	output_image0: Image = None,
	output_image1: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _onlyzero_overwrite_maximum_diamond as op

    return op(
        device=device,
		src=input_image,
		dst0=output_image0,
		dst1=output_image1
    )


@plugin_function
def power(
    input_image: Image,
	output_image: Image = None,
	scalar: float = 0,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _power as op

    return op(
        device=device,
		src=input_image,
		dst=output_image,
		scalar=float(scalar)
    )


@plugin_function
def power_images(
    input_image0: Image,
	input_image1: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _power_images as op

    return op(
        device=device,
		src0=input_image0,
		src1=input_image1,
		dst=output_image
    )


@plugin_function
def range(
    input_image: Image,
	output_image: Image = None,
	start_x: int = 0,
	stop_x: int = 0,
	step_x: int = 0,
	start_y: int = 0,
	stop_y: int = 0,
	step_y: int = 0,
	start_z: int = 0,
	stop_z: int = 0,
	step_z: int = 0,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _range as op

    return op(
        device=device,
		src=input_image,
		dst=output_image,
		start_x=int(start_x),
		stop_x=int(stop_x),
		step_x=int(step_x),
		start_y=int(start_y),
		stop_y=int(stop_y),
		step_y=int(step_y),
		start_z=int(start_z),
		stop_z=int(stop_z),
		step_z=int(step_z)
    )


@plugin_function
def read_intensities_from_positions(
    input_image: Image,
	list: Image = None,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _read_intensities_from_positions as op

    return op(
        device=device,
		src=input_image,
		list=list,
		dst=output_image
    )


@plugin_function
def replace_intensities(
    input_image0: Image,
	input_image1: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _replace_intensities as op

    return op(
        device=device,
		src0=input_image0,
		src1=input_image1,
		dst=output_image
    )


@plugin_function
def replace_intensity(
    input_image: Image,
	output_image: Image = None,
	scalar0: float = 0,
	scalar1: float = 0,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _replace_intensity as op

    return op(
        device=device,
		src=input_image,
		dst=output_image,
		scalar0=float(scalar0),
		scalar1=float(scalar1)
    )


@plugin_function
def maximum_sphere(
    input_image: Image,
	output_image: Image = None,
	radius_x: int = 0,
	radius_y: int = 0,
	radius_z: int = 0,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _maximum_sphere as op

    return op(
        device=device,
		src=input_image,
		dst=output_image,
		radius_x=int(radius_x),
		radius_y=int(radius_y),
		radius_z=int(radius_z)
    )


@plugin_function
def minimum_sphere(
    input_image: Image,
	output_image: Image = None,
	radius_x: int = 0,
	radius_y: int = 0,
	radius_z: int = 0,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _minimum_sphere as op

    return op(
        device=device,
		src=input_image,
		dst=output_image,
		radius_x=int(radius_x),
		radius_y=int(radius_y),
		radius_z=int(radius_z)
    )


@plugin_function
def multiply_matrix(
    input_image0: Image,
	input_image1: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _multiply_matrix as op

    return op(
        device=device,
		src0=input_image0,
		src1=input_image1,
		dst=output_image
    )


@plugin_function
def reciprocal(
    input_image: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _reciprocal as op

    return op(
        device=device,
		src=input_image,
		dst=output_image
    )


@plugin_function
def set(
    input_image: Image,
	scalar: float = 0,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _set as op

    return op(
        device=device,
		src=input_image,
		scalar=float(scalar)
    )


@plugin_function
def set_column(
    input_image: Image,
	column: int = 0,
	value: float = 0,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _set_column as op

    return op(
        device=device,
		src=input_image,
		column=int(column),
		value=float(value)
    )


@plugin_function
def set_image_borders(
    input_image: Image,
	value: float = 0,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _set_image_borders as op

    return op(
        device=device,
		src=input_image,
		value=float(value)
    )


@plugin_function
def set_plane(
    input_image: Image,
	plane: int = 0,
	value: float = 0,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _set_plane as op

    return op(
        device=device,
		src=input_image,
		plane=int(plane),
		value=float(value)
    )


@plugin_function
def set_ramp_x(
    input_image: Image,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _set_ramp_x as op

    return op(
        device=device,
		src=input_image
    )


@plugin_function
def set_ramp_y(
    input_image: Image,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _set_ramp_y as op

    return op(
        device=device,
		src=input_image
    )


@plugin_function
def set_ramp_z(
    input_image: Image,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _set_ramp_z as op

    return op(
        device=device,
		src=input_image
    )


@plugin_function
def set_row(
    input_image: Image,
	row: int = 0,
	value: float = 0,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _set_row as op

    return op(
        device=device,
		src=input_image,
		row=int(row),
		value=float(value)
    )


@plugin_function
def set_nonzero_pixels_to_pixelindex(
    input_image: Image,
	output_image: Image = None,
	offset: int = 0,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _set_nonzero_pixels_to_pixelindex as op

    return op(
        device=device,
		src=input_image,
		dst=output_image,
		offset=int(offset)
    )


@plugin_function
def set_where_x_equals_y(
    input_image: Image,
	value: float = 0,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _set_where_x_equals_y as op

    return op(
        device=device,
		src=input_image,
		value=float(value)
    )


@plugin_function
def set_where_x_greater_than_y(
    input_image: Image,
	value: float = 0,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _set_where_x_greater_than_y as op

    return op(
        device=device,
		src=input_image,
		value=float(value)
    )


@plugin_function
def set_where_x_smaller_than_y(
    input_image: Image,
	value: float = 0,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _set_where_x_smaller_than_y as op

    return op(
        device=device,
		src=input_image,
		value=float(value)
    )


@plugin_function
def sign(
    input_image: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _sign as op

    return op(
        device=device,
		src=input_image,
		dst=output_image
    )


@plugin_function
def smaller(
    input_image0: Image,
	input_image1: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _smaller as op

    return op(
        device=device,
		src0=input_image0,
		src1=input_image1,
		dst=output_image
    )


@plugin_function
def smaller_constant(
    input_image: Image,
	output_image: Image = None,
	scalar: float = 0,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _smaller_constant as op

    return op(
        device=device,
		src=input_image,
		dst=output_image,
		scalar=float(scalar)
    )


@plugin_function
def smaller_or_equal(
    input_image0: Image,
	input_image1: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _smaller_or_equal as op

    return op(
        device=device,
		src0=input_image0,
		src1=input_image1,
		dst=output_image
    )


@plugin_function
def smaller_or_equal_constant(
    input_image: Image,
	output_image: Image = None,
	scalar: float = 0,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _smaller_or_equal_constant as op

    return op(
        device=device,
		src=input_image,
		dst=output_image,
		scalar=float(scalar)
    )


@plugin_function
def sobel(
    input_image: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _sobel as op

    return op(
        device=device,
		src=input_image,
		dst=output_image
    )


@plugin_function
def square_root(
    input_image: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _square_root as op

    return op(
        device=device,
		src=input_image,
		dst=output_image
    )


@plugin_function
def subtract_image_from_scalar(
    input_image: Image,
	output_image: Image = None,
	scalar: float = 0,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _subtract_image_from_scalar as op

    return op(
        device=device,
		src=input_image,
		dst=output_image,
		scalar=float(scalar)
    )


@plugin_function
def sum_reduction_x(
    input_image: Image,
	output_image: Image = None,
	blocksize: int = 0,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _sum_reduction_x as op

    return op(
        device=device,
		src=input_image,
		dst=output_image,
		blocksize=int(blocksize)
    )


@plugin_function
def sum_x_projection(
    input_image: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _sum_x_projection as op

    return op(
        device=device,
		src=input_image,
		dst=output_image
    )


@plugin_function
def sum_y_projection(
    input_image: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _sum_y_projection as op

    return op(
        device=device,
		src=input_image,
		dst=output_image
    )


@plugin_function
def sum_z_projection(
    input_image: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _sum_z_projection as op

    return op(
        device=device,
		src=input_image,
		dst=output_image
    )


@plugin_function
def transpose_xy(
    input_image: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _transpose_xy as op

    return op(
        device=device,
		src=input_image,
		dst=output_image
    )


@plugin_function
def transpose_xz(
    input_image: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _transpose_xz as op

    return op(
        device=device,
		src=input_image,
		dst=output_image
    )


@plugin_function
def transpose_yz(
    input_image: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _transpose_yz as op

    return op(
        device=device,
		src=input_image,
		dst=output_image
    )


@plugin_function
def undefined_to_zero(
    input_image: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _undefined_to_zero as op

    return op(
        device=device,
		src=input_image,
		dst=output_image
    )


@plugin_function
def variance_box(
    input_image: Image,
	output_image: Image = None,
	radius_x: int = 0,
	radius_y: int = 0,
	radius_z: int = 0,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _variance_box as op

    return op(
        device=device,
		src=input_image,
		dst=output_image,
		radius_x=int(radius_x),
		radius_y=int(radius_y),
		radius_z=int(radius_z)
    )


@plugin_function
def variance_sphere(
    input_image: Image,
	output_image: Image = None,
	radius_x: int = 0,
	radius_y: int = 0,
	radius_z: int = 0,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _variance_sphere as op

    return op(
        device=device,
		src=input_image,
		dst=output_image,
		radius_x=int(radius_x),
		radius_y=int(radius_y),
		radius_z=int(radius_z)
    )


@plugin_function
def write_values_to_positions(
    input_image: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _write_values_to_positions as op

    return op(
        device=device,
		src=input_image,
		dst=output_image
    )




import numpy as np
from typing import Optional

cl_buffer_datatype_dict = {
    bool: "bool",
    np.uint8: "uchar",
    np.uint16: "ushort",
    np.uint32: "uint",
    np.uint64: "ulong",
    np.int8: "char",
    np.int16: "short",
    np.int32: "int",
    np.int64: "long",
    np.float32: "float",
    np.complex64: "cfloat_t",
    int: "int",
    float: "float",
    np.float64: "float",
}

_supported_numeric_types = tuple(cl_buffer_datatype_dict.keys())


def astype(self, dtype: type):
    if dtype not in _supported_numeric_types:
        raise ValueError(
            "dtype "
            + str(dtype)
            + " not supported. Use one of "
            + str(_supported_numeric_types)
        )
    if dtype == self.dtype:
        return self

    from ._tier1 import copy
    from ._memory import create_like

    result = create_like(self, dtype=dtype)
    copy(input_image=self, output_image=result)
    return result


def max(self, axis: Optional[int] = None, out=None):
    from ._tier2 import maximum_of_all_pixels
    from ._tier1 import maximum_x_projection
    from ._tier1 import maximum_y_projection
    from ._tier1 import maximum_z_projection

    if axis == 0:
        result = maximum_z_projection(self)
    elif axis == 1:
        result = maximum_y_projection(self)
    elif axis == 2:
        result = maximum_x_projection(self)
    elif axis is None:
        result = maximum_of_all_pixels(self)
    else:
        raise ValueError("Axis " + axis + " not supported")
    if out is not None:
        from ._memory import pull
        from ._array import Image

        if isinstance(out, Image):
            np.copyto(out, pull(result).astype(out.dtype))
        else:
            out = result
    return result


def min(self, axis: Optional[int] = None, out=None):
    from ._tier2 import minimum_of_all_pixels
    from ._tier1 import minimum_x_projection
    from ._tier1 import minimum_y_projection
    from ._tier1 import minimum_z_projection

    if axis == 0:
        result = minimum_z_projection(self)
    elif axis == 1:
        result = minimum_y_projection(self)
    elif axis == 2:
        result = minimum_x_projection(self)
    elif axis is None:
        result = minimum_of_all_pixels(self)
    else:
        raise ValueError("Axis " + axis + " not supported")
    if out is not None:
        from ._memory import pull
        from ._array import Image

        if isinstance(out, Image):
            np.copyto(out, pull(result).astype(out.dtype))
    return result


def sum(self, axis: Optional[int] = None, out=None):
    from ._tier2 import sum_of_all_pixels
    from ._tier1 import sum_x_projection
    from ._tier1 import sum_y_projection
    from ._tier1 import sum_z_projection

    if axis == 0:
        result = sum_z_projection(self)
    elif axis == 1:
        result = sum_y_projection(self)
    elif axis == 2:
        result = sum_x_projection(self)
    elif axis is None:
        result = sum_of_all_pixels(self)
    else:
        raise ValueError("Axis " + axis + " not supported")
    if out is not None:
        from ._memory import pull
        from ._array import Image

        if isinstance(out, Image):
            np.copyto(out, pull(result).astype(out.dtype))
    return result


def __iadd__(x1, x2):
    from ._tier1 import copy

    temp = copy(x1)
    if isinstance(x2, _supported_numeric_types):
        from ._tier1 import add_image_and_scalar

        return add_image_and_scalar(temp, x1, scalar=x2)
    else:
        from ._tier1 import add_images_weighted

        return add_images_weighted(temp, x2, x1, factor0=1, factor1=1)


def __sub__(x1, x2):
    if isinstance(x2, _supported_numeric_types):
        from ._tier1 import add_image_and_scalar

        return add_image_and_scalar(x1, scalar=-x2)
    else:
        from ._tier1 import add_images_weighted

        return add_images_weighted(x1, x2, factor0=1, factor1=-1)


def __div__(x1, x2):
    if isinstance(x2, _supported_numeric_types):
        from ._tier1 import multiply_image_and_scalar

        return multiply_image_and_scalar(x1, scalar=1.0 / x2)
    else:
        from ._tier1 import divide_images

        return divide_images(x1, x2)


def __truediv__(x1, x2):
    return x1.__div__(x2)


def __idiv__(x1, x2):
    from ._tier1 import copy

    temp = copy(x1)
    if isinstance(x2, _supported_numeric_types):
        from ._tier1 import multiply_image_and_scalar

        return multiply_image_and_scalar(temp, x1, scalar=1.0 / x2)
    else:
        from ._tier1 import divide_images

        return divide_images(temp, x2, x1)


def __itruediv__(x1, x2):
    return x1.__idiv__(x2)


def __mul__(x1, x2):
    if isinstance(x2, _supported_numeric_types):
        from ._tier1 import multiply_image_and_scalar

        return multiply_image_and_scalar(x1, scalar=x2)
    else:
        from ._tier1 import multiply_images

        return multiply_images(x1, x2)


def __imul__(x1, x2):
    from ._tier1 import copy

    temp = copy(x1)
    if isinstance(x2, _supported_numeric_types):
        from ._tier1 import multiply_image_and_scalar

        return multiply_image_and_scalar(temp, x1, scalar=x2)
    else:
        from ._tier1 import multiply_images

        return multiply_images(temp, x2, x1)


def __gt__(x1, x2):
    if isinstance(x2, _supported_numeric_types):
        from ._tier1 import greater_constant

        return greater_constant(x1, scalar=x2)
    else:
        from ._tier1 import greater

        return greater(x1, x2)


def __ge__(x1, x2):
    if isinstance(x2, _supported_numeric_types):
        from ._tier1 import greater_or_equal_constant

        return greater_or_equal_constant(x1, scalar=x2)
    else:
        from ._tier1 import greater_or_equal

        return greater_or_equal(x1, x2)


def __lt__(x1, x2):
    if isinstance(x2, _supported_numeric_types):
        from ._tier1 import smaller_constant

        return smaller_constant(x1, scalar=x2)
    else:
        from ._tier1 import smaller

        return smaller(x1, x2)


def __le__(x1, x2):
    if isinstance(x2, _supported_numeric_types):
        from ._tier1 import smaller_or_equal_constant

        return smaller_or_equal_constant(x1, scalar=x2)
    else:
        from ._tier1 import smaller_or_equal

        return smaller_or_equal(x1, x2)


def __eq__(x1, x2):
    if isinstance(x2, _supported_numeric_types):
        from ._tier1 import equal_constant

        return equal_constant(x1, scalar=x2)
    else:
        from ._tier1 import equal

        return equal(x1, x2)


def __ne__(x1, x2):
    if isinstance(x2, _supported_numeric_types):
        from ._tier1 import not_equal_constant

        return not_equal_constant(x1, scalar=x2)
    else:
        from ._tier1 import not_equal

        return not_equal(x1, x2)


def __pow__(x1, x2):
    if isinstance(x2, _supported_numeric_types):
        from ._tier1 import power

        return power(x1, exponent=x2)
    else:
        from ._tier1 import power_images

        return power_images(x1, x2)


def __ipow__(x1, x2):
    from ._tier1 import copy

    temp = copy(x1)
    if isinstance(x2, _supported_numeric_types):
        from ._tier1 import power

        return power(temp, x1, exponent=x2)
    else:
        from ._tier1 import power_images

        return power_images(temp, x2, x1)


def __iter__(self):
    class MyIterator:
        def __init__(self, image):
            self.image = image
            self._iter_index = 0

        def __next__(self):
            import numpy as np
            from ._memory import create
            from ._tier1 import copy_slice

            if not hasattr(self, "_iter_index"):
                self._iter_index = 0
            if self._iter_index < self.image.shape[0]:
                if len(self.image.shape) < 3:
                    result = np.asarray(self.image)[self._iter_index]
                elif len(self.image.shape) == 3:
                    output = create(self.image.shape[1:])
                    result = copy_slice(self.image, output, self._iter_index)
                else:
                    raise ValueError("Only 1D, 2D or 3D array are supported.")
                self._iter_index = self._iter_index + 1
                return result
            else:
                raise StopIteration

    return MyIterator(self)


def __getitem__(self, index):
    raise NotImplementedError("Not implemented yet.")


def __setitem__(self, index, value):
    raise NotImplementedError("Not implemented yet.")


# adapted from https://github.com/napari/napari/blob/d6bc683b019c4a3a3c6e936526e29bbd59cca2f4/napari/utils/notebook_display.py#L54-L73
def _plt_to_png(self):
    """PNG representation of the image object for IPython.
    Returns
    -------
    In memory binary stream containing a PNG matplotlib image.
    """
    import matplotlib.pyplot as plt
    from io import BytesIO

    with BytesIO() as file_obj:
        plt.savefig(file_obj, format="png")
        plt.close()  # supress plot output
        file_obj.seek(0)
        png = file_obj.read()
    return png


def _png_to_html(self, png):
    import base64

    url = "data:image/png;base64," + base64.b64encode(png).decode("utf-8")
    return f'<img src="{url}"></img>'


def _repr_html_(self):
    """HTML representation of the image object for IPython.
    Returns
    -------
    HTML text with the image and some properties.
    """
    import numpy as np
    import matplotlib.pyplot as plt
    from ._functionalities import imshow

    size_in_pixels = np.prod(self.size)
    size_in_bytes = size_in_pixels * self.dtype.itemsize

    labels = self.dtype == np.uint32

    # In case the image is 2D, 3D and larger than 100 pixels, turn on fancy view
    if len(self.shape) in (2, 3) and size_in_pixels >= 100:
        import matplotlib.pyplot as plt

        imshow(self, labels=labels, continue_drawing=True, colorbar=not labels)
        image = self._png_to_html(self._plt_to_png())
    else:
        return (
            "<pre>cle.array("
            + str(np.asarray(self))
            + ", dtype="
            + str(self.dtype)
            + ")</pre>"
        )

    units = ["B", "kB", "MB", "GB", "TB", "PB"]
    unit_index = 0
    while size_in_bytes > 1024 and unit_index < len(units) - 1:
        size_in_bytes /= 1024
        unit_index += 1
    size = "{:.1f}".format(size_in_bytes) + " " + units[unit_index]

    histogram = ""
    if size_in_bytes < 100 * 1024 * 1024:
        if not labels:
            from ._tier3 import histogram

            num_bins = 32
            h = np.asarray(
                histogram(self, nbins=num_bins, min=self.min(), max=self.max())
            )
            plt.figure(figsize=(1.8, 1.2))
            plt.bar(range(0, len(h)), h)
            # hide axis text
            # https://stackoverflow.com/questions/2176424/hiding-axis-text-in-matplotlib-plots
            # https://pythonguides.com/matplotlib-remove-tick-labels
            frame1 = plt.gca()
            frame1.axes.xaxis.set_ticklabels([])
            frame1.axes.yaxis.set_ticklabels([])
            plt.tick_params(left=False, bottom=False)
            histogram = self._png_to_html(self._plt_to_png())
        min_max = (
            "<tr><td>min</td><td>"
            + str(self.min())
            + "</td></tr>"
            + "<tr><td>max</td><td>"
            + str(self.max())
            + "</td></tr>"
        )
    else:
        min_max = ""
    all = [
        "<table>",
        "<tr>",
        "<td>",
        image,
        "</td>",
        '<td style="text-align: center; vertical-align: top;">',
        '<b><a href="https://github.com/clEsperanto/pyclesperanto_prototype" target="_blank">cle._</a> image</b><br/>',
        "<table>",
        "<tr><td>shape</td><td>"
        + str(self.shape).replace(" ", "&nbsp;")
        + "</td></tr>",
        "<tr><td>dtype</td><td>" + str(self.dtype) + "</td></tr>",
        "<tr><td>size</td><td>" + size + "</td></tr>",
        min_max,
        "</table>",
        histogram,
        "</td>",
        "</tr>",
        "</table>",
    ]
    return "\n".join(all)




# this code is auto-generated by the script 'pyclesperanto_autogen_tier_script.ipynb'.
# Do not edit manually. Instead, edit the script and run it again.

from ._core import Device
from ._array import Image
from ._decorators import plugin_function


@plugin_function
def masked_voronoi_labeling(
    input_image: Image,
	mask: Image = None,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _masked_voronoi_labeling as op

    return op(
        device=device,
		src=input_image,
		mask=mask,
		dst=output_image
    )


@plugin_function
def voronoi_labeling(
    input_image: Image,
	output_image: Image = None,
	device: Device = None
) -> Image:
    from ._pyclesperanto import _voronoi_labeling as op

    return op(
        device=device,
		src=input_image,
		dst=output_image
    )




# pyclesperanto version
VERSION_CODE = (0, 7, 0)
VERSION_STATUS = ""
VERSION = ".".join(str(x) for x in VERSION_CODE) + VERSION_STATUS

# clic version
CLIC_VERSION_CODE = (0, 7, 0)
CLIC_VERSION_STATUS = ""
CLIC_VERSION = ".".join(str(x) for x in CLIC_VERSION_CODE) + CLIC_VERSION_STATUS

COMMON_ALIAS = "cle"



import warnings

from ._array import Array, Image
from ._core import Device, get_device

import numpy as np
from typing import Tuple, Optional


def create(
    shape: Tuple[int, ...],
    dtype: Optional[type] = None,
    mtype: Optional[str] = None,
    device: Optional[Device] = None,
) -> Image:
    """Create a new image on the device.

    Parameters
    ----------
    shape : Tuple[int, ...]
        Shape of the image in (z,y,x)
    dtype : type, optional
        Data type of the image (np.int8, np.float32, etc.), np.float32 by default if None
    mtype : str, optional
        Memory type of the image (buffer, image), buffer by default if None
    device : Device, optional
        Device on which the image is created, current device by default if None

    Returns
    -------
    Image
        Created an empty image on the device
    """
    if not isinstance(shape, tuple):
        shape = tuple(shape)
    if dtype is None:
        dtype = np.float32
    if dtype in [float, np.float64]:
        dtype = np.float32
        warnings.warn(
            "Warning: float64 type is not a supported in GPUs. Casting data to float32 type.",
            UserWarning,
        )
    if mtype is None:
        mtype = "buffer"
    if device is None:
        device = get_device()
    return Array.create(shape, dtype, mtype, device)


def create_like(
    array: Image,
    dtype: Optional[type] = None,
    mtype: Optional[str] = None,
    device: Optional[Device] = None,
) -> Image:
    """Create a new image on the device with the same shape and dtype as the input image.

    Parameters
    ----------
    array : Image
        Input image
    dtype : type, optional
        Data type of the image (np.int8, np.float32, etc.), np.float32 by default if None
    mtype : str, optional
        Memory type of the image (buffer, image), buffer by default if None
    device : Device, optional
        Device on which the image is created, current device by default if None

    Returns
    -------
    Image
        Created an empty image on the device
    """
    if dtype is None:
        dtype = array.dtype
    if dtype in [float, np.float64]:
        dtype = np.float32
        warnings.warn(
            "Warning: float64 type is not a supported in GPUs. Casting data to float32 type.",
            UserWarning,
        )
    return create(array.shape, dtype, mtype, device)


def push(
    array: Image,
    dtype: Optional[type] = None,
    mtype: Optional[str] = None,
    device: Optional[Device] = None,
) -> Image:
    """Create a new image on the device and push the input image into it.

    Parameters
    ----------
    array : Image
        Input image
    dtype : type, optional
        Data type of the image (np.int8, np.float32, etc.), np.float32 by default if None
    mtype : str, optional
        Memory type of the image (buffer, image), buffer by default if None
    device : Device, optional
        Device on which the image is created, current device by default if None

    Returns
    -------
    Image
        Created image on the device with the input image data
    """
    if isinstance(array, Array):
        return array
    if array.dtype in [float, np.float64]:
        array = array.astype(np.float32)
    if dtype is None:
        dtype = array.dtype
    return create(array.shape, dtype, mtype, device).set(array)


def pull(array: Image) -> np.ndarray:
    """Pull the input image from the device to the host.

    Parameters
    ----------
    array : Image
        Input image

    Returns
    -------
    np.ndarray
        Image data
    """
    if isinstance(array, np.ndarray):
        return array
    return array.get()
